@@ -492,6 +492,9 @@ class TypeConverter(object):
         default = []
         other_tokens = []
 
+        if parts[0].name == '::':
+            parts = parts[1:]
+
         # Handle default (initial) values properly.
         for i, t in enumerate(parts):
             if t.name == '[' and arrayBegin == 0:
@@ -508,43 +511,40 @@ class TypeConverter(object):
         if arrayBegin < arrayEnd:
             parts = parts[:arrayBegin] + parts[arrayEnd + 1:]
 
-        name = None
-        if needs_name_removed and len(parts) > 1:
-            name = parts.pop().name
-
         modifiers = []
-        type_name = []
+        type_name = ['']
+        last_type = tokenize.SYNTAX
         templated_types = []
         i = 0
         end = len(parts)
         while i < end:
             p = parts[i]
-            if keywords.is_keyword(p.name):
+            if keywords.is_builtin_modifiers(p.name):
                 modifiers.append(p.name)
             elif p.name == '<':
                 templated_tokens, new_end = self._get_template_end(
                     parts, i + 1)
                 templated_types = self.to_type(templated_tokens)
                 i = new_end - 1
-                # Don't add a spurious :: to data members being initialized.
-                next_index = i + 1
-                if next_index < end and parts[next_index].name == '::':
-                    i += 1
-            elif p.name not in ('*', '&', '>'):
-                # Ensure that names have a space between them.
-                if (type_name and type_name[-1].token_type == tokenize.NAME and
-                        p.token_type == tokenize.NAME):
-                    type_name.append(
-                        tokenize.Token(
-                            tokenize.SYNTAX,
-                            ' ',
-                            0,
-                            0))
-                type_name.append(p)
+            elif p.name not in ('*', '&'):
+                if (
+                    last_type == tokenize.NAME and
+                    p.token_type == tokenize.NAME
+                ):
+                    type_name.append('')
+                type_name[-1] += p.name
+                last_type = p.token_type
             else:
                 other_tokens.append(p)
             i += 1
-        type_name = ''.join([t.name for t in type_name])
+
+        name = None
+        if len(type_name) == 1 or keywords.is_builtin_type(type_name[-1]):
+            needs_name_removed = False
+        if needs_name_removed:
+            name = type_name.pop()
+
+        type_name = ' '.join([t for t in type_name])
         return (name,
                 type_name,
                 templated_types,
@@ -558,7 +558,7 @@ class TypeConverter(object):
 
         result = []
         type_modifiers = []
-        pointer = reference = array = False
+        pointer = reference = False
         first_token = None
         default = []
 
@@ -569,27 +569,13 @@ class TypeConverter(object):
                 del default[0]  # Remove flag.
             end = type_modifiers[-1].end
 
-            needs_name_removed = True
-            if len(type_modifiers) == 1:
-                needs_name_removed = False
-            else:
-                last = type_modifiers[-1].name
-                second_to_last = type_modifiers[-2].name
-                if (
-                    last == '>' or
-                    keywords.is_builtin_type(last) or
-                    second_to_last == '::' or
-                    keywords.is_builtin_type_modifiers(second_to_last)
-                ):
-                    needs_name_removed = False
-
             (name, type_name, templated_types, modifiers,
              _, __) = self.declaration_to_parts(type_modifiers,
-                                                needs_name_removed)
+                                                True)
 
             parameter_type = Type(first_token.start, first_token.end,
                                   type_name, templated_types, modifiers,
-                                  reference, pointer, array)
+                                  reference, pointer, False)
             p = Parameter(first_token.start, end, name,
                           parameter_type, default)
             result.append(p)
@@ -612,7 +598,7 @@ class TypeConverter(object):
             if s.name == ',':
                 add_parameter()
                 type_modifiers = []
-                pointer = reference = array = False
+                pointer = reference = False
                 first_token = None
                 default = []
             elif default:
@@ -622,7 +608,7 @@ class TypeConverter(object):
             elif s.name == '&':
                 reference = True
             elif s.name == '[':
-                array = True
+                pointer = True
             elif s.name == ']':
                 pass  # Just don't add to type_modifiers.
             elif s.name == '=':
@@ -683,6 +669,7 @@ class ASTBuilder(object):
         self.visibility = visibility
         # Keep the state whether we are currently handling a typedef or not.
         self._handling_typedef = False
+        self._handling_const = False
         self.converter = TypeConverter(self.namespace_stack)
 
     def handle_error(self, msg, token):
@@ -747,14 +734,18 @@ class ASTBuilder(object):
                 # Fall through--handle like any other method.
 
             # Handle data or function declaration/definition.
-            syntax = tokenize.SYNTAX
             temp_tokens, last_token = \
-                self._get_var_tokens_up_to(syntax, '(', ';', '{', '}')
+                self._get_var_tokens_up_to(True, '(', ';', '{', '}')
             if last_token.name == '}':
                 return None
 
             temp_tokens.insert(0, token)
             if last_token.name == '(' or last_token.name == '{':
+                # Ignore static_assert
+                if temp_tokens[-1].name == 'static_assert':
+                    self._ignore_up_to(';')
+                    return None
+
                 # If there is an assignment before the paren,
                 # this is an expression, not a method.
                 for i, elt in reversed(list(enumerate(temp_tokens))):
@@ -764,7 +755,7 @@ class ASTBuilder(object):
                     ):
                         temp_tokens.append(last_token)
                         new_temp, last_token = \
-                            self._get_var_tokens_up_to(tokenize.SYNTAX, ';')
+                            self._get_var_tokens_up_to(False, ';')
                         temp_tokens.extend(new_temp)
                         break
 
@@ -786,13 +777,11 @@ class ASTBuilder(object):
             if last_token.name == '{':
                 assert_parse(temp_tokens, 'not enough tokens')
 
-                self._add_back_tokens(temp_tokens[1:])
                 self._add_back_token(last_token)
+                self._add_back_tokens(temp_tokens[1:])
                 method_name = temp_tokens[0].name
                 method = getattr(self, 'handle_' + method_name, None)
                 if not method:
-                    # Must be declaring a variable.
-                    # TODO(nnorwitz): handle the declaration.
                     return None
                 return method()
             return self._get_method(temp_tokens, 0, None, False)
@@ -849,34 +838,39 @@ class ASTBuilder(object):
                     if c == '(':
                         paren = i
                 if value.startswith('\\'):
-                    value = value[1:]
+                    value = value[1:].strip()
                 return Define(token.start, token.end, name, value)
             if name.startswith('undef'):
                 # Remove "undef".
                 name = name[5:].strip()
                 assert name
                 self.define.discard(name)
-            if name.startswith('if') and name[2:3].isspace():
-                condition = name[3:].strip()
-                if condition.startswith('0') or condition.startswith('(0)'):
-                    self._skip_if0blocks()
         return None
 
-    def _get_tokens_up_to(self, expected_token_type, expected_token):
-        return self._get_var_tokens_up_to(expected_token_type,
+    def _get_tokens_up_to(self, expected_token):
+        return self._get_var_tokens_up_to(False,
                                           expected_token)[0]
 
-    def _get_var_tokens_up_to(self, expected_token_type, *expected_tokens):
+    def _get_var_tokens_up_to(self, skip_bracket_content, *expected_tokens):
         last_token = self._get_next_token()
         tokens = []
-        count = 0
-        while (count != 0 or
-               last_token.token_type != expected_token_type or
+        count1 = 0
+        count2 = 0
+        while (count1 != 0 or
+               count2 != 0 or
+               last_token.token_type != tokenize.SYNTAX or
                last_token.name not in expected_tokens):
             if last_token.name == '[':
-                count += 1
+                count1 += 1
             elif last_token.name == ']':
-                count -= 1
+                count1 -= 1
+            if skip_bracket_content:
+                if last_token.name == 'operator':
+                    skip_bracket_content = False
+                elif last_token.name == '<':
+                    count2 += 1
+                elif last_token.name == '>':
+                    count2 -= 1
             tokens.append(last_token)
             temp_token = self._get_next_token()
             if temp_token.name == '(' and last_token.name in self.define:
@@ -886,23 +880,8 @@ class ASTBuilder(object):
             last_token = temp_token
         return tokens, last_token
 
-    def _ignore_up_to(self, token_type, token):
-        self._get_tokens_up_to(token_type, token)
-
-    def _skip_if0blocks(self):
-        count = 1
-        while True:
-            token = self._get_next_token()
-            if token.token_type != tokenize.PREPROCESSOR:
-                continue
-
-            name = token.name[1:].lstrip()
-            if name.startswith('endif'):
-                count -= 1
-                if count == 0:
-                    break
-            elif name.startswith('if'):
-                count += 1
+    def _ignore_up_to(self, token):
+        self._get_tokens_up_to(token)
 
     def _get_matching_char(self, open_paren, close_paren, get_next_token=None):
         if get_next_token is None:
@@ -943,10 +922,14 @@ class ASTBuilder(object):
 
     def get_name(self, seq=None):
         """Returns ([tokens], next_token_info)."""
-        get_next_token = self._get_next_token
         if seq is not None:
             it = iter(seq)
-            get_next_token = lambda: next(it)
+
+            def get_next_token():
+                return next(it)
+        else:
+            get_next_token = self._get_next_token
+
         next_token = get_next_token()
         tokens = []
         last_token_was_name = False
@@ -968,7 +951,7 @@ class ASTBuilder(object):
         return tokens, next_token
 
     def get_method(self, modifiers, templated_types):
-        return_type_and_name = self._get_tokens_up_to(tokenize.SYNTAX, '(')
+        return_type_and_name = self._get_tokens_up_to('(')
         assert len(return_type_and_name) >= 1
         return self._get_method(
             return_type_and_name, modifiers, templated_types,
@@ -989,6 +972,14 @@ class ASTBuilder(object):
             assert_parse(token.name == '(', token)
 
         name = return_type_and_name.pop()
+        if (len(return_type_and_name) > 1 and
+                return_type_and_name[-1].name == '>' and
+                (name.name == '>=' or name.name == '>')):
+            n = return_type_and_name.pop()
+            name = tokenize.Token(tokenize.SYNTAX,
+                                  n.name + name.name,
+                                  n.start, name.end)
+
         if (len(return_type_and_name) > 1 and
             (return_type_and_name[-1].name == 'operator' or
              return_type_and_name[-1].name == '~')):
@@ -1055,14 +1046,17 @@ class ASTBuilder(object):
                 modifiers |= FUNCTION_THROW
                 token = self._get_next_token()
                 assert_parse(token.name == '(', token)
-                # Consume everything between the (parens).
+                # Consume everything between the parens.
                 list(self._get_matching_char('(', ')'))
                 token = self._get_next_token()
             elif token.name == token.name.upper():
-                # HACK(nnorwitz):  assume that all upper-case names
-                # are some macro we aren't expanding.
+                # Assume that all upper-case names are some macro.
                 modifiers |= FUNCTION_UNKNOWN_ANNOTATION
                 token = self._get_next_token()
+                if token.name == '(':
+                    # Consume everything between the parens.
+                    list(self._get_matching_char('(', ')'))
+                    token = self._get_next_token()
             else:
                 self._add_back_token(token)
                 token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)
@@ -1071,7 +1065,7 @@ class ASTBuilder(object):
         if token.name == '&' or token.name == '&&':
             token = self._get_next_token()
 
-        if token.name == '}' or token.name == '#endif':
+        if token.name == '}' or token.token_type == tokenize.PREPROCESSOR:
             self._add_back_token(token)
             token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)
 
@@ -1156,14 +1150,8 @@ class ASTBuilder(object):
         # TODO(nnorwitz): if there is only One name like in the
         # example above, punt and assume the last bit is the class name.
 
-        # Ignore a :: prefix, if exists so we can find the first real name.
         i = 0
-        if token_seq[0].name == '::':
-            i = 1
-        # Ignore a :: suffix, if exists.
         end = len(token_seq) - 1
-        if token_seq[end - 1].name == '::':
-            end -= 1
 
         # Make a copy of the sequence so we can append a sentinel
         # value. This is required for get_name will has to have some
@@ -1180,8 +1168,6 @@ class ASTBuilder(object):
             names.append(new_name)
             i += len(new_name)
 
-        # Now that we have the names, it's time to undo what we did.
-
         # Remove the sentinel value.
         names[-1].pop()
         # Flatten the token sequence for the return type.
@@ -1247,12 +1233,12 @@ class ASTBuilder(object):
                 token = next_token
 
         if token.token_type == tokenize.SYNTAX and token.name == '(':
-            self._ignore_up_to(tokenize.SYNTAX, ')')
+            self._ignore_up_to(')')
             token = self._get_next_token()
 
         # Handle underlying type.
         if token.token_type == tokenize.SYNTAX and token.name == ':':
-            _, token = self._get_var_tokens_up_to(tokenize.SYNTAX, '{', ';')
+            _, token = self._get_var_tokens_up_to(False, '{', ';')
 
         # Handle forward declarations.
         if token.token_type == tokenize.SYNTAX and token.name == ';':
@@ -1344,7 +1330,11 @@ class ASTBuilder(object):
         pass
 
     def handle_const(self):
-        pass
+        self._handling_const = True
+        token = self._get_next_token()
+        result = self._generate_one(token)
+        self._handling_const = False
+        return result
 
     def handle_inline(self):
         pass
@@ -1365,7 +1355,7 @@ class ASTBuilder(object):
             return self.get_method(FUNCTION_VIRTUAL + FUNCTION_DTOR, None)
         assert_parse(token.token_type == tokenize.NAME or token.name == '::',
                      token)
-        return_type_and_name = self._get_tokens_up_to(tokenize.SYNTAX, '(')
+        return_type_and_name = self._get_tokens_up_to('(')
         return_type_and_name.insert(0, token)
         if token2 is not token:
             return_type_and_name.insert(1, token2)
@@ -1391,10 +1381,21 @@ class ASTBuilder(object):
         self.visibility = VISIBILITY_PRIVATE
 
     def handle_friend(self):
-        tokens = self._get_tokens_up_to(tokenize.SYNTAX, ';')
-        assert tokens
-        t0 = tokens[0]
-        return Friend(t0.start, t0.end, tokens, self.namespace_stack)
+        tokens, last = self._get_var_tokens_up_to(False, '(', ';')
+        if last.name == '(':
+            tokens.append(last)
+            self._add_back_tokens(tokens)
+            token = self._get_next_token()
+            if token.name == '::':
+                token = self._get_next_token()
+            result = self._generate_one(token)
+        else:
+            if tokens[0].name == 'class':
+                tokens = tokens[1:]
+            result = self.converter.to_type(tokens)[0]
+
+        assert result
+        return Friend(result.start, result.end, result, self.namespace_stack)
 
     def handle_static_cast(self):
         pass
@@ -1412,15 +1413,14 @@ class ASTBuilder(object):
         pass
 
     def handle_delete(self):
-        tokens = self._get_tokens_up_to(tokenize.SYNTAX, ';')
+        tokens = self._get_tokens_up_to(';')
         assert tokens
         return Delete(tokens[0].start, tokens[0].end, tokens)
 
     def handle_typedef(self):
         token = self._get_next_token()
         if (token.token_type == tokenize.NAME and
-                keywords.is_keyword(token.name)):
-            # Token must be struct/enum/union/class.
+                keywords.is_builtin_other_modifiers(token.name)):
             method = getattr(self, 'handle_' + token.name)
             self._handling_typedef = True
             tokens = [method()]
@@ -1429,7 +1429,7 @@ class ASTBuilder(object):
             tokens = [token]
 
         # Get the remainder of the typedef up to the semi-colon.
-        tokens.extend(self._get_tokens_up_to(tokenize.SYNTAX, ';'))
+        tokens.extend(self._get_tokens_up_to(';'))
 
         # TODO(nnorwitz): clean all this up.
         assert tokens
@@ -1470,7 +1470,7 @@ class ASTBuilder(object):
         while i < len_tokens:
             key = tokens[i].name
             i += 1
-            if keywords.is_keyword(key) or key == ',':
+            if keywords.is_keyword(key) or key == ',' or key == '.':
                 continue
             type_name = default = None
             if i < len_tokens:
@@ -1479,12 +1479,11 @@ class ASTBuilder(object):
                     assert_parse(i < len_tokens, '%s %s' % (i, tokens))
                     default, _ = self.get_name(tokens[i:])
                     i += len(default)
-                else:
-                    if tokens[i - 1].name != ',':
-                        # We got something like: Type variable.
-                        # Re-adjust the key (variable) and type_name (Type).
-                        key = tokens[i - 1].name
-                        type_name = tokens[i - 2]
+                elif tokens[i - 1].name != ',':
+                    # We got something like: Type variable.
+                    # Re-adjust the key (variable) and type_name (Type).
+                    key = tokens[i - 1].name
+                    type_name = tokens[i - 2]
 
             result[key] = (type_name, default)
         return result
@@ -1510,7 +1509,7 @@ class ASTBuilder(object):
             elif token.name == 'friend':
                 return self.handle_friend()
         self._add_back_token(token)
-        tokens, last = self._get_var_tokens_up_to(tokenize.SYNTAX, '(', ';')
+        tokens, last = self._get_var_tokens_up_to(False, '(', ';')
         tokens.append(last)
         self._add_back_tokens(tokens)
         if last.name == '(':
@@ -1546,9 +1545,6 @@ class ASTBuilder(object):
                 token = self._get_next_token()
                 if token.name != 'virtual':
                     self._add_back_token(token)
-                else:
-                    # TODO(nnorwitz): store that we got virtual for this base.
-                    pass
             base, next_token = self.get_name()
             bases_ast = self.converter.to_type(base)
             assert_parse(len(bases_ast) == 1, bases_ast)
@@ -1580,7 +1576,9 @@ class ASTBuilder(object):
             # Handle attribute.
             elif token.token_type == tokenize.NAME:
                 self._add_back_token(token)
-                name_tokens, token = self.get_name()
+                attribute, token = self.get_name()
+                if len(attribute) > 1 or attribute[0].name != 'final':
+                    name_tokens = attribute
             class_name = ''.join([t.name for t in name_tokens])
         bases = None
         if token.token_type == tokenize.SYNTAX:
@@ -1609,7 +1607,8 @@ class ASTBuilder(object):
 
         body = None
         if token.token_type == tokenize.SYNTAX and token.name == '{':
-            ast = ASTBuilder(self.get_scope(), self.filename, class_name,
+            name = class_name or '__unamed__'
+            ast = ASTBuilder(self.get_scope(), self.filename, name,
                              visibility, self.namespace_stack,
                              quiet=self.quiet)
             body = list(ast.generate())
@@ -1624,7 +1623,7 @@ class ASTBuilder(object):
                                            class_name, bases, None,
                                            body, self.namespace_stack)
 
-                    modifiers = []
+                    modifiers = ['const'] if self._handling_const else []
                     return self._create_variable(class_token,
                                                  token.name, new_class,
                                                  modifiers, token.name)
@@ -1662,7 +1661,7 @@ class ASTBuilder(object):
         return None
 
     def handle_using(self):
-        tokens = self._get_tokens_up_to(tokenize.SYNTAX, ';')
+        tokens = self._get_tokens_up_to(';')
         assert tokens
         return Using(tokens[0].start, tokens[0].end, tokens)
 
@@ -1701,11 +1700,11 @@ class ASTBuilder(object):
         pass
 
     def handle_return(self):
-        tokens = self._get_tokens_up_to(tokenize.SYNTAX, ';')
+        tokens = self._get_tokens_up_to(';')
         return Return(tokens[0].start, tokens[0].end, tokens)
 
     def handle_goto(self):
-        tokens = self._get_tokens_up_to(tokenize.SYNTAX, ';')
+        tokens = self._get_tokens_up_to(';')
         assert_parse(len(tokens) == 1, tokens)
         return Goto(tokens[0].start, tokens[0].end, tokens[0].name)
 
@@ -1728,10 +1727,10 @@ class ASTBuilder(object):
         pass
 
     def handle_break(self):
-        self._ignore_up_to(tokenize.SYNTAX, ';')
+        self._ignore_up_to(';')
 
     def handle_continue(self):
-        self._ignore_up_to(tokenize.SYNTAX, ';')
+        self._ignore_up_to(';')
 
 
 def builder_from_source(source, filename, quiet=False):
@@ -1745,7 +1744,7 @@ def builder_from_source(source, filename, quiet=False):
       ASTBuilder
 
     """
-    return ASTBuilder(tokenize.get_tokens(source, quiet=quiet),
+    return ASTBuilder(tokenize.get_tokens(source),
                       filename,
                       quiet=quiet)
 
