@@ -145,22 +145,22 @@ class WarningHunter(object):
             msg = "unable to find '{}'".format(filename)
             self._add_warning(msg, node)
         else:
+            ast_list = None
             try:
                 builder = ast.builder_from_source(source, filename,
                                                   quiet=self.quiet)
-                module = Module(filename,
-                                [_f for _f in builder.generate() if _f])
-            except (ast.ParseError,
-                    tokenize.TokenError) as error:
+                ast_list = [_f for _f in builder.generate() if _f]
+            except tokenize.TokenError:
+                pass
+            except ast.ParseError as error:
                 if not self.quiet:
                     print(
                         "Exception while processing '{}': {}".format(
                             filename,
                             error),
                         file=sys.stderr)
-                module = Module(filename, None)
-            else:
-                self._update_symbol_table(module)
+            module = Module(filename, ast_list)
+            self._update_symbol_table(module)
         return module
 
     def _read_and_parse_includes(self):
@@ -375,9 +375,11 @@ class WarningHunter(object):
                         _add_template_use('<typedef>', alias.templated_types,
                                           node.namespace)
                 elif isinstance(node, ast.Friend):
-                    if node.expr and node.expr[0].name == 'class':
-                        name = ''.join([n.name for n in node.expr[1:]])
-                        _add_reference(name, node.namespace)
+                    expr = node.expr
+                    if isinstance(expr, ast.Type):
+                        _add_reference(expr.name, node.namespace)
+                    elif isinstance(expr, ast.Function):
+                        _process_function(expr)
                 elif isinstance(node, ast.Class) and node.body is not None:
                     if node.body:
                         ast_seq.append(node.body)
