@@ -19,7 +19,6 @@ from __future__ import absolute_import
 from __future__ import print_function
 from __future__ import unicode_literals
 
-import sys
 
 __author__ = 'nnorwitz@google.com (Neal Norwitz)'
 
@@ -102,7 +101,7 @@ def _get_char(source, start, i):
     return i + 1 if i != -1 else start + 1
 
 
-def get_tokens(source, quiet=False):
+def get_tokens(source):
     """Returns a sequence of Tokens.
 
     Args:
@@ -121,8 +120,7 @@ def get_tokens(source, quiet=False):
     int_or_float_digits = INT_OR_FLOAT_DIGITS
     int_or_float_digits2 = int_or_float_digits | set('.')
 
-    # Only ignore errors while in a #if 0 block.
-    ignore_errors = False
+    # Ignore tokens while in a #if 0 block.
     count_ifs = 0
 
     i = 0
@@ -159,7 +157,8 @@ def get_tokens(source, quiet=False):
             token_type = SYNTAX
             i += 1
             new_ch = source[i]
-            if new_ch == c:
+            # Do not merge '>>' or '>>=' into a single token
+            if new_ch == c and c != '>':
                 i += 1
                 new_ch = source[i]
             if new_ch == '=':
@@ -219,10 +218,11 @@ def get_tokens(source, quiet=False):
         elif c == '#':                           # Find pre-processor command.
             token_type = PREPROCESSOR
             got_if = source[i:i + 3] == '#if'
-            if ignore_errors and source[i:i + 6] == '#endif':
+            if count_ifs and source[i:i + 6] == '#endif':
                 count_ifs -= 1
                 if count_ifs == 0:
-                    ignore_errors = False
+                    source = source[:i].ljust(i + 6) + source[i + 6:]
+                    continue
 
             # Handle preprocessor statements (\ continuations).
             while True:
@@ -251,31 +251,36 @@ def get_tokens(source, quiet=False):
                     continue
 
                 if got_if:
-                    condition = source[start + 4:i].lstrip()
+                    begin = source.find('(', start, i)
+                    if begin == -1:
+                        begin = source.find(' ', start)
+                    begin = begin + 1
+                    s1 = source.find(' ', begin)
+                    s2 = source.find(')', begin)
+                    s3 = source.find('\n', begin)
+                    s = min([x for x in (s1, s2, s3, end) if x != -1])
+
+                    condition = source[begin:s]
                     if (
-                        ignore_errors or
-                        condition.startswith('0') or
-                        condition.startswith('(0)')
+                        count_ifs or
+                        condition == '0' or
+                        condition == '__OBJC__'
                     ):
                         count_ifs += 1
-                        ignore_errors = True
                 break
         elif c == '\\':                          # Handle \ in code.
             # This is different from the pre-processor \ handling.
             i += 1
             continue
-        elif ignore_errors:
+        elif count_ifs:
             # Ignore bogus code when we are inside an #if block.
             i += 1
             continue
         else:
-            if not quiet:
-                sys.stderr.write(
-                    'Got invalid token in %s @ %d token:%s: %r\n' %
-                    ('?', i, c, source[i - 10:i + 10]))
-            raise TokenError('unexpected token')
+            raise TokenError('unexpected token %s' % c)
 
-        if i <= 0:
-            print('Invalid index, exiting now.')
-            return
+        if count_ifs:
+            continue
+
+        assert i > 0
         yield Token(token_type, source[start:i], start, i)
